/*
 * disasm - a MIPS disassembler.
 *
 * Copyright (C) 1995-1996 Kai Harrekilde-Petersen <khp@dolphinics.no>
 *
 * Version: 0.38
 * Date: 22. August 1996
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>

#include "disasm.h"

#define VERSION "0.38"

char *mips_regs[32] = {
  "zero", "at", "v0", "v1",
  "a0",   "a1", "a2", "a3",
  "t0",   "t1", "t2", "t3",
  "t4",   "t5", "t6", "t7",
  "s0",   "s1", "s2", "s3",
  "s4",   "s5", "s6", "s7",
  "t8",   "t9", "k0", "k1",
  "gp",   "sp", "s8", "ra"
};

char *algor_regs[32] = {
  "$zero", "$at", "$v0", "$v1",
  "$a0",   "$a1", "$a2", "$a3",
  "$t0",   "$t1", "$t2", "$t3",
  "$t4",   "$t5", "$t6", "$t7",
  "$s0",   "$s1", "$s2", "$s3",
  "$s4",   "$s5", "$s6", "$s7",
  "$t8",   "$t9", "$k0", "$k1",
  "$gp",   "$sp", "$s8", "$ra"
};

/* Variables needed by the disassembly functions */
char **regs;
char **CP0regs;
isa_t isa;
int debug;

inline int
sign_extend(int imm)
{
#if 0
  return (imm & (1 << 15)) ? (~0<<16) | imm : imm;
#else
  /* I guess this aught to generate optimum code on all platforms */
  return (int) ((short) imm);
#endif
}

void
pr_bug(char *string, unsigned insn)
{
  sprintf(string,".word\t0x%08x\t# ***BUG***",insn);
  return;
}

void
pr_print(char *string, insn_st **class, unsigned insn)
{
  /* stub function */
  return;
}

void
pr_isa(char *string, insn_st **class, unsigned insn)
{
  /* stub function */
  return;
}

void
pr_inv(char *string, unsigned insn)
{
  sprintf(string,".word\t0x%08x\t# invalid instruction",insn);
  return;
}

void
pr_res(char *string, unsigned insn)
{
  sprintf(string,".word\t0x%08x\t# Reserved instruction (causes "
	  "exception)",insn);
  return;
}

void
pr_none(char *string, unsigned insn)
{
  /* This shouldn't happen */
  sprintf(string,".word\t0x%08x\t# Whoops! - internal error.",insn);
  return;
}

void
pr_nop(char *string, insn_st *insn)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s",insn->name);
  } else {
    sprintf(string,"%s\t# (mips-%i insn)",insn->name,insn->isa);
  }
  return;
}  


#if 0
void
pr_special(void)
{
  /* TYPE_SPECIAL isn't a printed type as such; TYPE_SPECIAL means
   * that we should look in the [5..0] field for the sub-opcode
   */
  return;
}
#endif

void
pr_regimm(char *string, unsigned insn, unsigned addr)
{
  unsigned r1, r2;
  unsigned imm16;

  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  imm16 = FIELD(insn,0,16);

  switch(regimm[r2].type) {
  case TYPE_BRA:
    pr_bra(string,&regimm[r2],r1,addr,imm16);
    break;
  case TYPE_TRAPI:
    pr_trapi(string,&regimm[r2],r1,imm16);
    break;
  case TYPE_RES:
    pr_res(string,insn);
    break;
  case TYPE_INV:
    pr_inv(string,insn);
    break;
  case TYPE_PRINT:
    /* Ugh, right. Need to do something intelligent here. */
    pr_bra(string,&regimm[r2],r1,addr,imm16); /* FIXME */
    break;
  default:
    pr_bug(string,insn);
    break;
  } /* switch(regimm[r2].type) */
}

void
pr_immu(char *string, insn_st *insn, int r1, int r2, unsigned imm)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s,%s,0x%04x",insn->name,regs[r2],regs[r1],imm);
  } else {
    sprintf(string,"%s\t%s,%s,0x%04x\t# (mips-%i insn)",insn->name,
	    regs[r2],regs[r1],imm,insn->isa); 
  }
  return;
}  

void
pr_reg(char *string, insn_st *insn, int r1, int r2, int r3)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s,%s,%s", insn->name,
	    regs[r3],regs[r1],regs[r2]);
  } else {
    sprintf(string,"%s\t%s,%s,%s\t# (mips-%i insn)", insn->name,
	    regs[r3],regs[r1],regs[r2],insn->isa);
  }
  return;
}

void
pr_jump(char *string, insn_st *insn, unsigned addr, unsigned offset)
{
  unsigned long target = (addr & (~0 << 28)) | (offset << 2);
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%08lx",insn->name,target);
  } else {
    sprintf(string,"%s\t%08lx\t# (mips-%i insn)",insn->name,target,insn->isa);
  }
  return;
}

void
pr_imm(char *string, insn_st *insn, int r1, int r2, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm >= 0 ) {
      sprintf(string,"%s\t%s,%s,0x%04x",insn->name,
	      regs[r2],regs[r1],imm);
    } else {
      sprintf(string,"%s\t%s,%s,%i",insn->name,
	      regs[r2],regs[r1],imm);
    }
  } else {
    if(imm >= 0 ) {
      sprintf(string,"%s\t%s,%s,0x%04x\t# (mips-%i insn)",insn->name,
	      regs[r2],regs[r1],imm,insn->isa);
    } else {
      sprintf(string,"%s\t%s,%s,%i\t# (mips-%i insn)",insn->name,
	      regs[r2],regs[r1],imm,insn->isa);
    }
  }
  return;
}

void
pr_bra(char *string, insn_st *insn, int r1, unsigned addr,
       unsigned imm)
{
  unsigned long target = addr + 4 + (imm << 2);
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s,%08lx",insn->name,regs[r1],target);
  } else {
    sprintf(string,"%s\t%s,%08lx\t# (mips-%i insn)",insn->name,regs[r1],
	    target,insn->isa);
  }
  return;
}

void
pr_ldst(char *string, insn_st *insn, int r1, int r2, int offset)
{
  offset = sign_extend(offset);
  if(isa>=insn->isa) {
    if(offset>=0) {
      sprintf(string,"%s\t%s,0x%x(%s)",insn->name,regs[r2],offset,regs[r1]);
    } else {
      sprintf(string,"%s\t%s,%i(%s)",insn->name,regs[r2],offset,regs[r1]);
    }
  } else {
    if(offset>=0) {
      sprintf(string,"%s\t%s,0x%x(%s)\t# (mips-%i insn)",insn->name,
	      regs[r2],offset,regs[r1],insn->isa);
    } else {
      sprintf(string,"%s\t%s,%i(%s)\t# (mips-%i insn)",insn->name,regs[r2],
	      offset,regs[r1],insn->isa);
    }
  }
  return;
}

void
pr_shift(char *string, insn_st *insn, int r2, int r3, int shift)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s,%s,0x%02x",insn->name,regs[r3],regs[r2],shift);
  } else {
    sprintf(string,"%s\t%s,%s,0x%02x\t# (mips-%i insn)",insn->name,
	    regs[r3],regs[r2],shift,insn->isa);
  }
  return;
}

void
pr_jr(char *string, insn_st *insn, int reg)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s",insn->name,regs[reg]);
  } else {
    sprintf(string,"%s\t%s\t# (mips-%i insn)",insn->name,regs[reg],insn->isa);
  }
  return;
}

void
pr_syscall(char *string, insn_st *insn, unsigned imm)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t0x%05x",insn->name,imm);
  } else {
    sprintf(string,"%s\t0x%05x\t# (mips-%i insn)",insn->name,imm,insn->isa);
  }
  return;
}

void
pr_hilo(char *string, insn_st *insn, int reg)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s",insn->name,regs[reg]);
  } else {
    sprintf(string,"%s\t%s\t# (mips-%i insn)",insn->name,regs[reg],insn->isa);
  }
  return;
}

void
pr_move(char *string, insn_st *insn, int r1, int r2)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s\t%s,%s",insn->name,regs[r1],regs[r2]);
  } else {
    sprintf(string,"%s\t%s,%s\t# (mips-%i insn)",insn->name,regs[r1],
	    regs[r2],insn->isa);
  }
  return;
}

void
pr_cp(void)
{
  /* I guess the CPn stuff need a /bit/ more handling :-) */
  return;
}

void
pr_cache(char *string, insn_st *insn, int r1, int r2, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm < 0) {
      sprintf(string,"%s\t0x%x,%i(%s)",insn->name,r2,imm,regs[r1]);
    } else {
      sprintf(string,"%s\t0x%x,0x%x(%s)",insn->name,r2,imm,regs[r1]);
    }
  } else {
    if(imm < 0) {
      sprintf(string,"%s\t0x%x,%i(%s)\t# (mips-%i insn)",insn->name,r2,imm,
	      regs[r1],insn->isa);
    } else {
      sprintf(string,"%s\t0x%x,0x%x(%s)\t (mips-%i insn)",insn->name,r2,imm,
	      regs[r1],insn->isa);
    }
  }
  return;
}

void
pr_trapi(char *string, insn_st *insn, int reg, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm>=0) {
      sprintf(string,"%s\t%s,0x%04x",insn->name,regs[reg],imm);
    } else {
      sprintf(string,"%s\t%s,%i",insn->name,regs[reg],imm);
    }
  } else {
    if(imm>=0) {
      sprintf(string,"%s\t%s,0x%04x\t# (mips-%i insn)",insn->name,
	      regs[reg],imm,insn->isa);
    } else {
      sprintf(string,"%s\t%s,%i\t# (mips-%i insn)",insn->name,regs[reg],
	      imm,insn->isa);
    }
  }
  return;
}

void
print_special(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;

  signed int imm;

  /*
   * Well, the *real* opcode field lies at [31..26], but we look at
   * the sub-opcode here (main opcode is "special").
   */
  opcode = FIELD(insn,0,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  shift = FIELD(insn,6,5);

  switch(special[opcode].type) {
  case TYPE_REG:
    pr_reg(string,&special[opcode],r1,r2,r3);
    break;
  case TYPE_IMM:
    /* convert imm16 into a "true" signed int */
    imm = sign_extend(imm16);
    if(imm >= 0 ) {
      sprintf(string,"%s\t%s,%s,0x%04x",special[opcode].name,
	      regs[r2],regs[r1],imm);
    } else {
      sprintf(string,"%s\t%s,%s,%i",special[opcode].name,
	      regs[r2],regs[r1],imm);
    }
    break;
  case TYPE_SHIFT:
    sprintf(string,"%s\t%s,%s,0x%02x",special[opcode].name,regs[r3],
	    regs[r2],shift);
    break;
  case TYPE_JR:
    sprintf(string,"%s\t%s",special[opcode].name,regs[r1]);
    break;
  case TYPE_SYSCALL:
    /* extract "code" field */
    imm = FIELD(insn,6,20);
    sprintf(string,"%s\t0x%05x",special[opcode].name,imm);
    break;
  case TYPE_NOP:
    sprintf(string,"%s",special[opcode].name);
    break;
  case TYPE_HILO:
    sprintf(string,"%s\t%s",special[opcode].name,regs[r2]);
    break;
  case TYPE_MOVE:
    sprintf(string,"%s\t%s,%s",opcodes[opcode].name,regs[r1],regs[r2]);
    break;
  case TYPE_NONE:
    /* illegal insn */
    sprintf(string,".word\t0x%08x\t# illegal or reserved "
	    "instruction",insn);
    if(debug) {
      sprintf(string+strlen(string),"\n\t# in print_special():"
	      "switch(opcodes[].type)");
    }
    break;
  default:
    /* Internal error -- cannot happen! */
    sprintf(string,"Oops!\tinternal error in print_special().");
    break;
  }
}


/*
 * parse a COPz insn
 */
void
print_cp(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned cp;
  unsigned r1, r2, r3, r4;

  /*
   * Well, the true opcode field lies at [31..26], but we have
   * already decode that, and now turns to the sub-opcode at
   * [5..0] (main opcode is "COPz").
   */
  opcode = FIELD(insn,0,6);
  cp = FIELD(insn,26,2);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  r4 = FIELD(insn,6,5);

  switch(cp) {
  case 0: /* System co-proc */
    /* First decide if it is a tlbxx/eret/wait type insn */
    if(FIELD(insn,25,1)) {
      /* Yup, it is.  Look up the insn in the CP0_opcode[] */
      switch(CP0_opcode[opcode].type) {
      case TYPE_NOP:
	sprintf(string,"%s",CP0_opcode[opcode].name);
	break;
      case TYPE_NONE:
	sprintf(string,".word\t0x%08x\t# illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"\n\t# print_cp():CPzrs(%i):"
		  "TYPE_NONE",cp);
	}
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(CP0_opcode[opcode].type) */
    } else { /* if(!FIELD(insn,25,1)) */
      switch(COPzrs[r1].type) {
      case TYPE_CP:
	sprintf(string,"%sc0\t%s,%s",COPzrs[r1].name,
		regs[r2],CP0regs[r3]);
	break;
      default:
	sprintf(string,".word\t0x%08x\t# illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"illegal\t# print_cp():CPzrs(%i):"
		  "default",cp);
	}
	break;
      } /* switch(COPzrs) */
    } /* if(FIELD(insn,25,1)) */
    break;
    /* END OF CP0 PROCESSING */
  case 1: /* Floating Point co-proc */
    if(opcode == 0x00) {
      /* "standard" move from/to CP1 insns */
      switch(COPzrs[r1].type) {
      case TYPE_CP:
	sprintf(string,"%sc1\t%s,f%i",COPzrs[r1].name,
		regs[r2],r3);
	break;
      case TYPE_NOP:
	/* Naughty, naughty:  --FIXME!!!! */ 
	if(r1 == 16 || r1 == 17) {
	  sprintf(string,"%s.%s\tf%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r3,r2);
	} else {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\tf%i,f%i",CP1_func[opcode].name,
		  r3,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,".word\t0x%08x",insn);
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(COPzrs[r1].type) */
    } else {
      /* "real" FPU insns */
      switch(CP1_func[opcode].type) {
      case TYPE_REG:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\tf%i,f%i,f%i",
		  CP1_func[opcode].name,r4,r3,r2);
	} else {
	  sprintf(string,"%s.%s\tf%i,f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r4,r3,r2);
	}
	break;
      case TYPE_MOVE:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\tf%i,f%i",CP1_func[opcode].name,
		  r4,r3);
	} else {
	  sprintf(string,"%s.%s\tf%i,f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r4,r3,r2);
	}
	break;
      case TYPE_BRA:
	/* This is *compares* not branches(!) */
	/* C.cond.fmt fA,fB */
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_RES) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill\tf%i,f%i",CP1_func[opcode].name,
		  r3,r2);
	} else {
	  sprintf(string,"%s.%s\tf%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r3,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,".word\t0x%08x\t# illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"\n\t# in print_cp():"
		  "CP1_func:TYPE_NONE");
	}
	break;
      default:
	sprintf(string,"\t# Oops! internal error in print_cp():CP1_func)");
	break;
      } /* switch(CP1_func[opcode].type) */
    } /* !if(opcode == 0x00) */
    break;
    /* END OF CP1 PROCESSING */
  case 2: /* ??? co-proc */
    sprintf(string,".word\t0x%08x\t# cp%i: giving up!",insn,cp);
    break;
  case 3: /* invalid?? co-proc */
    sprintf(string,".word\t0x%08x\t# cp%i: illegal coproc",insn,cp);
    break;
  default: /* cannot happen */
    sprintf(string,"print_cp(): This cannot happen. cp = %i",cp);
    break;
  }
  return;
}

/*
 * examine the machine instruction `insn' and format it into `string',
 * under the assumption that it resides at `addr'.
 */
void
disassemble(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;
  unsigned offset;
  signed int imm;
  unsigned long target;

  opcode = FIELD(insn,26,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  offset = FIELD(insn,0,6);
  shift = FIELD(insn,6,6);

  /*
   * Handle all "special cases" first
   */
  switch(opcodes[opcode].type) {
  case TYPE_PRINT:
    pr_bug(string,insn);
    break;
  case TYPE_SPECIAL:
    /* "sll zero,zero,0" => "nop" */
    if(insn == 0x00) {
      sprintf(string,"nop");
      return;
    }
    /* map "or rA,zero,rB" => "move rA,rB" */
    if(offset == 0x25 && r2 == 0x00) {
      sprintf(string,"move\t%s,%s",regs[r3],regs[r1]);
      return;
    }
    break;
  case TYPE_REGIMM:
#if 0 /* comment this out so I can see the full ghastness of it */
    if(regimm[r2].type == TYPE_PRINT && r1 == 0) {
      fixup_t *tmp_insn;
      int flag=0;
      target = (unsigned long) (addr + 4 + (sign_extend(imm16)<<2));
      i=0;
      do {
	tmp_insn = fixup_insns[i];
	if(tmp_insn.type==TYPE_REGIMM && tmp_insn.opcode==r2) {
	  flag=1;
	  /* match */
	  if(tmp_insn.decode==TYPE_BRA) {
	    sprintf(string,"%s\t%08lx",tmp_insn.name,target);
	  } else {
	    pr_bug(insn);
	  }
	}
	i++;
      } while(fixup_insn[i].name!=NULL && !flag);
      if!(flag) pr_bug(insn);
    } /* if(regimm[r2].type == TYPE_BRA && r1 == 0) */
#endif
    break;
  case TYPE_IMMU:
    /* "lui rA,zero,imm" => "lui rA,imm" */
    if(opcode == 0x0f && r1 == 0x00) {
      sprintf(string,"lui\t%s,0x%04x",regs[r2],imm16);
      return;
    }
    break;
  case TYPE_IMM:
    /* "addiu rA,zero,imm" => "li rA,imm" */
    if(opcode == 0x09 && r1 == 0x00) {
      imm = sign_extend(imm16);
      if(imm>=0) {
	sprintf(string,"li\t%s,0x%04x",regs[r2],imm);
      } else {
	sprintf(string,"li\t%s,%i",regs[r2],imm);
      }
      return;
    }
    break;
  case TYPE_BRA:
    /* "beq zero,zero,target" => "b target" */
    if(r1 == 0 && r2 == 0) {
      /* shift imm16 two bits left and signed extend it */
      target = (unsigned long)(addr + 4 + (sign_extend(imm16) << 2));
      sprintf(string,"b\t%08lx",target);
      return;
    }
    /* "bCC r1,zero,target" => "bCCz r1,target" */
    if(r1 == 0 || r2 == 0) {
      /* compute target address */
      target = (unsigned long)(addr + 4 + (sign_extend(imm16)<<2));
      r1 += r2; /* Nice hack which cover both cases graciously */
      sprintf(string,"%sz\t%s,%08lx",opcodes[opcode].name,
	      regs[r1],target);
      return;
    }
    break;
  case TYPE_CP:
    if(COPzrs[r1].type == TYPE_CP && FIELD(insn,26,2) == 0
       && (r1 == 0x02 || r1 == 0x06)) {
      /* CFCz and CTCz are illegal for COP0 */
      sprintf(string,".word\t=x%08x\t# illegal or reserved "
	      "instruction",insn);
      return;
    }
    break;
  default:
    /* It is perfectly valid to take the default action. This just means that
     * the insn will be handled by the "normal case" code below */
    break;
  } /* Handle "special cases" */

  /*
   * Now handle the normal cases
   */
  switch(opcodes[opcode].type) {
  case TYPE_SPECIAL:
    print_special(string,insn,addr);
    break;
  case TYPE_REGIMM:
    pr_regimm(string,insn,addr);
    break;
  case TYPE_IMMU:
    sprintf(string,"%s\t%s,%s,0x%04x",opcodes[opcode].name,regs[r2],
	    regs[r1],imm16);
    break;
  case TYPE_JUMP:
    target = (addr & (~0 << 28)) | (imm26 << 2);
    sprintf(string,"%s\t%08lx",opcodes[opcode].name,target);
    break;
  case TYPE_IMM:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm >= 0) {
      sprintf(string,"%s\t%s,%s,0x%04x",opcodes[opcode].name,regs[r2],
	      regs[r1],imm);
    } else {
      sprintf(string,"%s\t%s,%s,%i",opcodes[opcode].name,regs[r2],regs[r1],
	      imm);
    }
    break;
  case TYPE_BRA:
    /* shift imm16 two bits left and signed extend it */
    imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
    /* compute target address */
    target = (unsigned long)(addr + 4 + imm);
    sprintf(string,"%s\t%s,%s,%08lx",opcodes[opcode].name,regs[r1],regs[r2],
	    target);
    break;
  case TYPE_LDST:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    sprintf(string,"%s\t%s,0x%x(%s)",opcodes[opcode].name,regs[r2],imm,
	    regs[r1]);
    break;
  case TYPE_JR:
    sprintf(string,"%s\t%s",opcodes[opcode].name,regs[r1]);
    break;
  case TYPE_CP:
    print_cp(string,insn,addr);
    break;
  case TYPE_CACHE:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm < 0) {
      sprintf(string,"%s\t0x%x,%i(%s)",opcodes[opcode].name,r2,imm,regs[r1]);
    } else {
      sprintf(string,"%s\t0x%x,0x%x(%s)",opcodes[opcode].name,r2,imm,
	      regs[r1]);
    }
    break;
  case TYPE_NONE:
    sprintf(string,".word\t0x%08x\t# illegal or reserved "
	    "instruction",insn);
    if(debug) {
      sprintf(string+strlen(string),"\n\t# in disassemble():Normal "
	      "cases:TYPE_NONE");
    }
    break;
  case TYPE_NOP:
  case TYPE_REG:
  case TYPE_SHIFT:
  case TYPE_SYSCALL:
  case TYPE_HILO:
  case TYPE_MOVE:
    /* These can only happen as a sub-opcode under SPECIAL */
  default:
    /* illegal insn or something wrong */
    sprintf(string,"Oops!\tinternal error in disassemble(): "
	    "opcode type = %03x",opcodes[opcode].type );
    break;
  }
}

/*
 * convert a big/little endian word into a little/big endian word
 */
inline unsigned long
nuxi (const unsigned long value)
{
  return ((value >> 24) & 0x000000ffUL)
       | ((value >>  8) & 0x0000ff00UL)
       | ((value <<  8) & 0x00ff0000UL)
       | ((value << 24) & 0xff000000UL);
}

void
main(int argc, char **argv)
{
  u_long insn;
  char stri[1025];
  char *binname;
  int tmp;
  int flag;
  int reverse;
  unsigned long addr;
  int counter;
  FILE *infile;
  FILE *outfile;
  struct stat file_s;

  /*
   * DEFAULTS:
   * Not debugging.
   * ISA = MIPS-III
   * No little/big endian conversion
   * Start address = 0xbfc0000
   * Emit MIPS style reg names (rx, as opposed to Algorithmics style: $rx)
   * output file: stdout.
   */
  debug = 0;
  isa = mips3;
  reverse = 0;
  addr = 0xbfc00000;
  regs = mips_regs;
  outfile = stdout;

  /* find exactly the "basename" of argv[0] */
  binname = strrchr(argv[0],'/');
  if(binname == NULL) {
    binname = argv[0];
  } else {
    binname++;
  }

  if(argc == 1) {
    fprintf(stderr,"%s : [-dhprvV] [ -m <ISA>] [-a <start address>] "
	    "[-o <outfile>] <filename>\n",binname);
    exit(0);
  }

  /* initialize optind & optarg for getopt() */
  tmp = 0;
  do {
    int option_idx = 0;
    static struct option long_opts[] = {
      {"address", 1, 0, 'a'},
      {"debug", 0, 0, 'd'},
      {"help", 0, 0, 'h'},
      {"mips",1, 0, 'm'},
      {"output", 1, 0, 'o'},
      {"algor-regs", 0, 0, 'p'},
      {"reverse", 0, 0, 'r'},
      {"version", 0, 0, 'v'},
      {"copying", 0, 0, 'V'},
      {0, 0, 0, 0}
    };
    flag = getopt_long(argc,argv,"a:dhm:o:prvV",
		       long_opts, &option_idx);
    switch(flag) {
    case 'a':
      addr = strtoul(optarg,(char **)NULL,0);
      break;
    case 'd':
      debug = 1;
      break;
    case 'h':
      fprintf(stderr,
	     "%s: [-dhprvV] [-m<ISA>] [-a start-address] "
	      "[-o outfile] filename\n  [--address=start-address]"
	      " [--mips=ISA-level] [--debug] [--help]\n"
	      "  [--algor-regs]  [--reverse] [--version] [--copying]"
	      " [--output=out-file]\n",binname);
      exit(0);
      break;
    case 'm':
      tmp = atof(optarg);
      if(tmp>0 && tmp<4) isa = (isa_t) tmp;
      break;
    case 'o':
      if(strcmp(optarg,"-")) outfile = fopen(optarg,"w");
      break;
    case 'p':
      regs = algor_regs;
      break;
    case 'r': 
      reverse = 1;
      break;
    case 'v':
      fprintf(stderr,"%s version " VERSION "\n",binname);
      exit(0);
      break;
    case 'V':
      fprintf(stderr,"%s - a MIPS disassembler, version " VERSION "\n"
	     "Copyright (C) 1995-1996 Kai Harrekilde-Petersen\n\n"
	     "This is free software; see the source for copying conditions.\n"
	     "The program is distributed in the hope that it will be useful,\n"
	     "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	     "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n",
	     binname);
      exit(0);
      break;
    case ':':
    case '?':
      printf("I'm sorry. I don't understand `%s'\n",argv[optind]);
      exit(2);
    }
  } while(flag != -1);

  /* Make the CP0 register name ptr point at the right place */
  switch (isa) {
  case mips1:
    CP0regs = r3k_CP0regs;
    break;
  case mips2:
  case mips3:
    CP0regs = r4k_CP0regs;
    break;
  default:
    /* This cannot happen */
  }

  /* Probe the file. */
  if(optind < argc) {
    flag=stat(argv[optind],&file_s);
    if (flag==-1) {
      fprintf(stderr,"Couldn't stat file %s, aborting program !!\n",
	      argv[optind]);
      exit(1);
    }
  } else {
    puts("Hmmm. I couldn't find a filename in there.");
    exit(0);
  }

/* Opening file */
  infile=fopen(argv[optind],"r");
  strcpy(stri,argv[optind]);

  for(counter = 0; counter < file_s.st_size/4; counter++) {
    fread(&insn,sizeof(unsigned long),1,infile);
    if(reverse) insn = nuxi(insn);
    stri[0] = '\0'; /* clear the string buffer */
    disassemble(stri,insn,addr);
    fprintf(outfile,"    %08lx: %08lx\t%s\n",(unsigned long) addr,insn,stri);
    addr += 4;
  }
  fclose(infile);
  fclose(outfile);
  exit(0);
}
