/*
 * disasm - a MIPS disassembler.
 *
 * Copyright (C) 1996 Kai Harrekilde-Petersen <khp@dolphinics.no>
 *
 * File: disasm.h
 * Version: 0.38
 *
 * Date: 22. August 1996
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __DISASM_H__
#define __DISASM_H__

#define FIELD(insn,shift,width) ((insn >> shift) & (~((~0) << width)))

typedef enum {
  r3000 = 3,
  r4000 = 4
} cpu_t;

typedef enum {
  mips0 = 0, /* MIPS-0   ISA (A mythical and nonexistant ISA) */
  mips1 = 1, /* MIPS-I   ISA */
  mips2 = 2, /* MIPS-II  ISA */
  mips3 = 3  /* MIPS-III ISA */
} isa_t;

/* instruction classes */
typedef enum {
  TYPE_PRINT = -5, /* This insn is printed differently under magic circumstances. */
  TYPE_ISA = -4, /* Some opcodes change meaning between the ISAs. curse 'em! */
  TYPE_INV = -3,
  TYPE_RES = -2,
  TYPE_NONE = -1, /* opcode is invalid (unused?) */
  TYPE_NOP = 0,
  TYPE_IMMU = 1,
  TYPE_REG = 2,
  TYPE_JUMP = 3,
  TYPE_IMM = 4,
  TYPE_BRA = 5,
  TYPE_LDST = 6,
  TYPE_SPECIAL = 7,
  TYPE_REGIMM = 8,
  TYPE_SHIFT = 9,
  TYPE_JR = 10,
  TYPE_SYSCALL = 11,
  TYPE_HILO = 12,
  TYPE_MOVE = 13,
  TYPE_CP = 14,
  TYPE_CACHE = 15,
  TYPE_TRAPI = 16
} insn_e;

typedef struct {
  char *name;
  insn_e type;
  isa_t isa;
} insn_st;

/*
 * Instruction types:
 * PRINT:	kludge to detect rfe/eret so we can do the Right Thing
 * INV:		invalid but does not generate a Reserved Insn exception
 * RES:		opcode generates a Reserved Insn exception
 * NONE:	Illegal/Reserved insn.  General bad-boy.
 * NOP:		nop
 * IMMU:	andi	r1,r2,unsigned16
 * REG:		or	r1,r2,r3
 * JUMP:	j	imm28
 * IMM:		addi	r1,r2,signed16
 * BRA:		beq	r1,r2,`signed16 << 2'
 * LDST:	lw	r1,signed16(r2)
 * REGIMM:	ori	r1,r2,signed16 
 * SHIFT:	sll	r1,r2,unsigned16
 * JR:		jr	r1
 * SYSCALL:	syscall	unsigned20
 * HILO:	mtlo	r1
 * MOVE:	move	r1,r2
 * SPECIAL:	any of: SHIFT, REG, JR, SYSCALL, HILO, and MOVE
 * CP:		any of: NOP, REG
 * CACHE:	cache	imm5,signed16(r1) [ONLY R4x00]
 * TRAPI:	tnei	rt,signed16
 */

/*
 * This struct is used to identify insns that need special attention.
 * These are noted with type TYPE_OOB or TYPE_ISA in the main tables.
 * Usually, this is things like "sll $0,$0,0 -> nop" or
 * "bgezal $0,foo(bar) -> bal foo(bar)" etc.
 * When looking in the table, we first match the type, then the opcode.
 * When finding a TYPE_ISA in the main table, this means that the opcode
 * has changed meaning from ISA to ISA (curse them).  Specifically, this
 * is used in the rfe/eret case.
 */
typedef struct {
  char *name;
  insn_e type;
  insn_e decode;
  isa_t isa;
  unsigned opcode;
} fixup_t;

/*
 * Things defined in r3kinsn.c
 */
extern insn_st opcodes[];
extern insn_st special[];
extern insn_st regimm[];
extern insn_st COPzrs[];
extern insn_st COPzrt[];
extern insn_st CP0_opcode[];
extern insn_st CP1_func[];
extern insn_st CP1_fmt[];

extern char *r3k_CP0regs[];
extern char *r4k_CP0regs[];

extern fixup_t fixup_insns[];

extern void pr_bug(char *str, unsigned insn);
extern void pr_print(char *str, insn_st **class, unsigned insn);
extern void pr_isa(char *str, insn_st **class, unsigned insn);
extern void pr_inv(char *str, unsigned insn);
extern void pr_res(char *str, unsigned insn);
extern void pr_none(char *str, unsigned insn);
extern void pr_nop(char *str, insn_st *insn);
extern void pr_regimm(char *str, unsigned insn, unsigned addr);
extern void pr_immu(char *str, insn_st *insn, int r1, int r2, unsigned imm);
extern void pr_reg(char *str, insn_st *insn, int r1, int r2, int r3);
extern void pr_jump(char *str, insn_st *insn, unsigned addr, unsigned offset);
extern void pr_imm(char *str, insn_st *insn, int r1, int r2, int imm);
extern void pr_bra(char *str, insn_st *insn, int r1, unsigned addr, unsigned imm);
extern void pr_ldst(char *str, insn_st *insn, int r1, int r2, int offset);
extern void pr_shift(char *str, insn_st *insn, int r2, int r3, int shift);
extern void pr_jr(char *str, insn_st *insn, int reg);
extern void pr_syscall(char *str, insn_st *insn, unsigned imm);
extern void pr_hilo(char *str, insn_st *insn, int reg);
extern void pr_move(char *str, insn_st *insn, int r1, int r2);
extern void pr_cache(char *str, insn_st *insn, int r1, int r2, int imm);
extern void pr_trapi(char *str, insn_st *insn, int reg, int imm);

extern void print_special(char *string, unsigned insn, unsigned addr);
extern void print_cp(char *str, unsigned insn, unsigned addr);

#endif /* __DISASM_H__ */
